import os
from gzip import GzipFile
import contextlib
import tempfile
import subprocess
import glob
import pandas as pd
import numpy as np

AMINO_ACID_MAP = {'ALA':'A','ARG':'R','ASN':'N','ASP':'D','CYS':'C','GLN':'Q','GLU':'E', \
         'GLY':'G','HIS':'H','ILE':'I','LEU':'L','LYS':'K','MET':'M','PHE':'F',\
         'PRO':'P','SER':'S','THR':'T','TRP':'W','TYR':'Y','VAL':'V','PYX':'C',\
         'SEP':'S','TPO':'T','TYS':'Y','MK8':'L','M3L':'K','DPR':'P','DSN':'S',\
         'ALN':'A','DLY':'K','MVA':'V','MLE':'L','DLE':'L','CSO':'C','PTR':'Y',\
         'BMT':'T','DAL':'A','FAK':'K','MSE':'M','SMF':'A','HYP':'P'}

@contextlib.contextmanager
def gunzipper(gz_file: str):
    """
    Gunzipper: takes in a file name. Returns a valid file object in an unzipped form. Uses
    context manager to create a temporary file and save into that format. This
    was written for programmatic extraction of PDB files using Prody.

    Parameters:
    -----------
    gz_file [ required ]
        gzipped file
        type: str

    Returns:
    --------
    temporary file
    tempfile.NamedTemporaryFile
    """
    with tempfile.NamedTemporaryFile('r', suffix=os.path.splitext(gz_file)[1]) as temp_file:
        subprocess.check_call("gzip -dc {} >> {}".format(gz_file, temp_file.name), executable='/bin/bash', shell=True)
        yield temp_file

def add_corrected_fdr(results_df: pd.DataFrame, thresh_num: float = 0.25):
    """
    Creates corrected FDR for clumps-ptm results.
    Due to the sparsity of PTM modifications possible on a given
    protein, we perform FDR on the subset of structures that can
    yield a theoretical p-value < threshold

    Parameters:
    -----------
    results_df [ required ]
        results dataframe from clumps-ptm
        type: pd.DataFrame

    thresh_num [ required ]
        threshold for minimum theoretical p-value to consider for FDR
        type: float

    Returns:
    --------
    results_df with additional columns for corrected FDR
    pd.DataFrame

    """
    from scipy.special import comb
    from statsmodels.stats.multitest import multipletests

    corr_df = list()

    for sam in np.unique(results_df['clumpsptm_sampler']):
        _df_sam = results_df[results_df['clumpsptm_sampler']==sam].copy()

        _df_sam['fdr_max_pval'] = _df_sam.apply(lambda row: 1/comb(row['clumpsptm_sample_n'], row['clumpsptm_input_n']), 1)
        _df_sam['fdr_pass'] = _df_sam['fdr_max_pval'] <= thresh_num
        _df_sam['fdr_corr'] = 1
        _,_df_sam.loc[_df_sam['fdr_pass'],'fdr_corr'],_,_ = multipletests(
            _df_sam[_df_sam['fdr_pass']]['clumpsptm_pval'], method='fdr_bh',
        )

        corr_df.append(_df_sam)

    return pd.concat(corr_df)

def generate_clumpsptm_output(
    output_dir: str,
    protein_id: str = 'accession_number',
    site_id: str = 'variableSites',
    min_sites: int = 3,
    thresh_min_pval: bool = True,
    alphafold: bool = False
    ) -> pd.DataFrame:
    """
    Generate CLUMPS-PTM Output file.

    Parameters:
    -----------
    output_dir [ required ]
        output directory from clumps-ptm to pull results from
        type: str

    protein_id [ optional ]
        protein id used in input
        type: str

    site_id [ optional ]
        site id used in input
        type: str

    min_sites [ optional ]
        min-sites required for a given "clump"
        type: int

    thresh_min_pval [ optional ]
        whether to threshold empirical pvalues = 0 (due to insufficient permutations) to 1/n_permutations^e-1
        type: bool

    Returns:
    --------
    results dataframe
    pd.DataFrame

    """
    def _collapse_sites(df):
        """Collapser per feature type."""
        from statsmodels.stats.multitest import multipletests

        if alphafold:
            acc_df = df[
                ['geneSymbol','uniprot',protein_id,'clumpsptm_sampler',
                 'clumpsptm_niter','clumpsptm_pval','clumpsptm_input','clumpsptm_sample']
            ].set_index(protein_id).drop_duplicates().sort_values('clumpsptm_pval')
        else:
            acc_df = df[
                ['geneSymbol','uniprot',protein_id,'pdb','chain','clumpsptm_sampler',
                 'clumpsptm_niter','clumpsptm_pval','clumpsptm_input','clumpsptm_sample']
            ].set_index(protein_id).drop_duplicates().sort_values('clumpsptm_pval')

        site_collapse_df = df.reset_index()[
            [protein_id,site_id,'acc_res','acc_res_i',
             'uniprot_res_i','uniprot_res','pdb_res_i','pdb_res','value']
            ].groupby(protein_id).agg(list)

        acc_df = acc_df.join(site_collapse_df)

        # Post-processing
        acc_df['clumpsptm_input_n'] = acc_df['clumpsptm_input'].apply(lambda x: len(x.split("+")))
        acc_df['clumpsptm_sample_n'] = acc_df['clumpsptm_sample'].apply(lambda x: len(x.split("+")))

        # Erroneous
        acc_df = acc_df[acc_df['clumpsptm_input_n']!=acc_df['clumpsptm_sample_n']]

        # Protein - Gene
        acc_df["y"] = acc_df.index + " | " + acc_df['geneSymbol']

        # Filter out for min sites
        acc_df = acc_df[acc_df['clumpsptm_input'].apply(lambda x: len(x.split("+")) >= min_sites)]

        # Threshold lower end of p-values
        if thresh_min_pval:
            acc_df.loc[acc_df['clumpsptm_pval']==0,'clumpsptm_pval'] = 1/acc_df['clumpsptm_niter'][acc_df['clumpsptm_pval']==0]/10

        # FDR
        if acc_df.shape[0] == 0:
            return None

        _,acc_df['clumpsptm_fdr'],_,_ = multipletests(acc_df['clumpsptm_pval'].astype(float), method='fdr_bh', alpha=0.1)

        # Corrected FDR
        acc_df = add_corrected_fdr(acc_df)

        return acc_df

    ptm_files = glob.glob(os.path.join(output_dir, "*_ptm_clumpsptm*"))
    phosph_files = glob.glob(os.path.join(output_dir, "*_phosphoproteome_clumpsptm*"))
    acetyl_files = glob.glob(os.path.join(output_dir, "*_acetylome_clumpsptm*"))

    res_df = list()
    if len(ptm_files) > 0:
        clumps_ptm_df = pd.concat([pd.read_parquet(x) for x in ptm_files])
        clumps_ptm_df['clumpsptm_sampler'] = "ptm"
        clumps_ptm_df = _collapse_sites(clumps_ptm_df)
        res_df.append(clumps_ptm_df)

    if len(phosph_files) > 0:
        clumps_phosph_df = pd.concat([pd.read_parquet(x) for x in phosph_files])
        clumps_phosph_df['clumpsptm_sampler'] = "phosphoproteome"
        clumps_phosph_df = _collapse_sites(clumps_phosph_df)
        res_df.append(clumps_phosph_df)

    if len(acetyl_files) > 0:
        clumps_acetyl_df = pd.concat([pd.read_parquet(x) for x in acetyl_files])
        clumps_acetyl_df['clumpsptm_sampler'] = "acetylome"
        clumps_acetyl_df = _collapse_sites(clumps_acetyl_df)
        res_df.append(clumps_acetyl_df)

    if len(res_df) == 0:
        raise ValueError('NO RESULTS FILES FOUND.')

    return pd.concat(res_df)



import os
import sys
from typing import Union
import pandas as pd
from tqdm import tqdm

def buildPymol(
    pdb: str,
    chain: str,
    session_name: str,
    phosph_residues: Union[None, list] = None,
    acetyl_residues: Union[None, list] = None,
    chain_color: str = 'palecyan',
    phosph_color: str = 'deeppurple',
    acetyl_color: str = 'hotpink'
    ):
    """
    Build Pymol Session
    """
    import __main__
    import tempfile
    import os

    # Quiet and no GUI
    __main__.pymol_argv = [ 'pymol', '-qc']

    import pymol

    temp_name = next(tempfile._get_candidate_names())+'.cif'
    obj_name = "{}{}".format(pdb, chain)

    pymol.finish_launching()

    pymol.cmd.delete("all")
    pymol.cmd.fetch(obj_name, file=temp_name)

    # Stylings
    pymol.cmd.set('ray_trace_fog','0')
    pymol.cmd.set('ray_shadows','0')
    pymol.cmd.set('depth_cue','0')
    pymol.cmd.bg_color('white')
    pymol.cmd.set('antialias','4')
    pymol.cmd.set('cartoon_transparency','0')
    pymol.cmd.set('transparency','0.7')
    pymol.cmd.set('ray_trace_mode','3')
    pymol.cmd.color(chain_color, obj_name)
    pymol.cmd.show('surface', obj_name)
    pymol.cmd.remove('solvent')

    if phosph_residues is not None:
        pymol.cmd.show('spheres', 'resi {}'.format('+'.join(phosph_residues)))
        pymol.cmd.color(phosph_color, 'resi {}'.format('+'.join(phosph_residues)))

    if acetyl_residues is not None:
        pymol.cmd.show('spheres', 'resi {}'.format('+'.join(acetyl_residues)))
        pymol.cmd.color(acetyl_color, 'resi {}'.format('+'.join(acetyl_residues)))

    pymol.cmd.save(session_name+'.pse')
    os.remove(temp_name)

def buildPymol_from_result(entry: pd.Series, out_dir: Union[None, str] = None, include_idx_in_name: bool = False):
    """
    Build Pymol from entry.
    """
    import ast

    # Name of Pymol Session
    session_name = list()
    if include_idx_in_name:
        session_name.append(entry.name)

    session_name.append(entry['geneSymbol'])
    session_name.append(entry['clumpsptm_sampler'])

    if 'id' in entry:
        session_name.append(entry['id'])

    session_name = '_'.join(session_name)

    # Make directory
    if out_dir is not None:
        os.makedirs(out_dir, exist_ok=True)
        session_name = os.path.join(out_dir, session_name)

    phosph_residues = list()
    acetyl_residues = list()

    if entry['clumpsptm_sampler'] == 'phosphoproteome':
        phosph_residues = entry['clumpsptm_input'].split('+')
    elif entry['clumpsptm_sampler'] == 'acetylome':
        acetyl_residues = entry['clumpsptm_input'].split('+')
    elif entry['clumpsptm_sampler'] == 'ptm':
        for (res,num) in zip(ast.literal_eval(entry['pdb_res']), entry['clumpsptm_input'].split('+')):
            if res=='K':
                acetyl_residues.append(str(num))
            else:
                phosph_residues.append(str(num))

    if len(acetyl_residues)==0:
        acetyl_residues = None
    if len(phosph_residues)==0:
        phosph_residues = None

    buildPymol(
        entry['pdb'],
        entry['chain'],
        session_name,
        phosph_residues = phosph_residues,
        acetyl_residues = acetyl_residues
    )

def create_pymols_from_result(results_df: pd.DataFrame, out_dir: str = '.', pval_thresh: float = 0.1, include_idx_in_name: bool = False, verbose: bool = False):
    """Create pymol sessions for all results."""
    _df = results_df[results_df['clumpsptm_pval']<pval_thresh]

    if out_dir!='.':
        if verbose:
            print("   * writing out session files to {}".format(out_dir))

    for idx,row in tqdm(_df.iterrows(), total=_df.shape[0]):
        buildPymol_from_result(row, out_dir=out_dir, include_idx_in_name=include_idx_in_name)



import importlib,sys
import os
import pandas as pd
import numpy as np
from collections import defaultdict
from typing import Union
from scipy.spatial.distance import euclidean
from .utils import AMINO_ACID_MAP, gunzipper
from Bio import pairwise2
import subprocess
import glob
from agutil.parallel import parallelize2

class PdbStore(object):
    """
    Protein Data Bank Direcotry Store
    ------------------
    This object is used to keep track of protein data bank files
    that are downloaded in a reference directory.
    """
    def __init__(self, path):
        """
        Args:
            * path: PDB directory

        """
        self.pdb_root_dir = path
        self.pdb_dir = os.path.join(path, 'ftp.wwpdb.org/pub/pdb/data/structures/divided/pdb/')

    def __str__(self):
        return "PDBStore\n   * {} PDB files downloaded\n   * Directory Path: {}".format(len(self.downloaded_pdbs), self.pdb_root_dir)

    @property
    def downloaded_pdbs(self):
        """
        Get Downloaded Structures.
        ------------------
        Iterates through pdb structure direcotry to find what
        pdbs are already downloaded. Returns a set of downloaded PDBs.
        """
        try:
            return {
                    i[3:7]
                    for j in os.listdir(self.pdb_dir)
                    for i in os.listdir(os.path.join(self.pdb_dir, j))
                    if i.endswith('.ent.gz') and os.path.getsize(os.path.join(self.pdb_dir, j, i)) > 0
                   }
        except:
            return set()

    def missing_pdbs(self, pdbs: set):
        """
        Missing PDBs
        ------------------
        Set difference of input pdbs and all downloaded pdbs.

        Args:
            * pdbs: set of PDBs
        """
        return pdbs - self.downloaded_pdbs

    def download_missing_pdbs(self, pdbs_to_download: Union[set,list], n_threads=15):
        """
        Download Missing PDBs
        ------------------
        Args:
            * pdbs_to_download: a list or set of PDBs to download
            * n_threads: number of threads to use for downloads
        """
        if isinstance(pdbs_to_download, list):
            pdbs_to_download = set(pdbs_to_download)

        out_dir = self.pdb_root_dir

        @parallelize2(maximum=n_threads)
        def dl_pdb(pdb):
            """Download pdb."""
            try:
                cmd = "wget -qr ftp://ftp.wwpdb.org/pub/pdb/data/structures/divided/pdb/{}/pdb{}.ent.gz -P {}".format(pdb[1:3], pdb, out_dir)
                output = subprocess.check_call(cmd, executable='/bin/bash', shell=True)
                return None
            except:
                return pdb

        print("   * Downloading {} pdbs using {} threads".format(len(pdbs_to_download), n_threads))

        tmp = [dl_pdb(pdb) for pdb in pdbs_to_download]
        pdb_err = {callback() for callback in tmp}

        print("   * Downloaded {} / {} successfully.".format(len(pdbs_to_download-pdb_err) , len(pdbs_to_download)))

        return pdb_err

    def load(self, pdb, chain=None):
        """
        Load residues - amino acids.

        Returns:
            * PDBStream
        """
        from prody import parsePDBStream

        pdb_file = os.path.join(self.pdb_dir, pdb[1:3], "pdb{}.ent.gz".format(pdb))

        with gunzipper(pdb_file) as pfile:
            aa = parsePDBStream(pfile, chain=chain)

        return aa

    def load_header(self, pdb, chain=None):
        """
        Load PDB header

        Returns:
            * dict
        """
        from prody import parsePDBHeader

        pdb_file = os.path.join(self.pdb_dir, pdb[1:3], "pdb{}.ent.gz".format(pdb))
        return parsePDBHeader(pdb_file)

    def load_rd(self, pdb, chain=None):
        """
        Load residue dict
        """
        aa = self.load(pdb,chain)
        res_map = dict(zip(aa.getResnums(),aa.getResnames()))
        return {k:v for k,v in res_map.items() if v in AMINO_ACID_MAP}

    def load_dm(self, pdb, chain=None, point='centroid', return_centroid_coord=False):
        """
        Load distance matrix for pdb.
        """
        from scipy.spatial.distance import euclidean
        import numpy as np

        aa = self.load(pdb,chain=chain)

        xx = aa.getResnums()
        yy = aa.getCoords()
        zz = aa.getResnames()

        pdb_resids = {}
        for i in range(len(xx)):
            if zz[i] in AMINO_ACID_MAP:
                pdb_resids[xx[i]] = True

        mapped_pdb_to_aa = defaultdict(set)
        for idx,resnum in enumerate(xx):
            if resnum in pdb_resids:
                mapped_pdb_to_aa[resnum].add(zz[idx])

        coords = {}
        for i in range(len(xx)):
            if xx[i] not in pdb_resids:
                continue
            if xx[i] not in coords:
                coords[xx[i]] = []
            coords[xx[i]].append(yy[i])  ## add coordinates of an atom belonging to this residue

        ## Euclidean distance matrix
        D = []
        for i in range(len(pdb_resids)):
            D.append(np.zeros(i, dtype=np.float32))

        centroids = {}
        for k in coords:
            centroids[k] = np.mean(np.array(coords[k]), 0)

        co = [centroids[i] for i in pdb_resids]  ## pdb residue coordinates

        for i in range(len(pdb_resids)):
            for j in range(i):
                D[i][j] = euclidean(co[i], co[j])

        if return_centroid_coord:
            return (D, pdb_resids, mapped_pdb_to_aa, co)
        else:
            return (D, pdb_resids, mapped_pdb_to_aa)

class AlphaStore(object):
    """
    Alpha Fold Data Bank Directory Store
    ------------------
    This object is used to keep track of protein data bank files
    that are downloaded in a reference directory.
    """
    def __init__(self, path):
        """
        Args:
            * path: PDB directory

        """
        self.pdb_dir = path
        self.uniprot_dict = {x.split("-")[1]:x for x in glob.glob(os.path.join(self.pdb_dir, "*.pdb.gz"))}

        try:
            self.tar_file = glob.glob(os.path.join(self.pdb_dir, "*.tar"))[0]
            self.version = self.tar_file.split(".tar")[0].split("_")[-1]
        except:
            pass

    def __str__(self):
        return "AlphaStore\n   * {} PDB files downloaded\n   * Directory Path: {}".format(len(self.uniprots), self.pdb_dir)

    @property
    def uniprots(self):
        """
        Return downloaded uniprot models.
        """
        return list(self.uniprot_dict.keys())

    def load(self, uniprot):
        """
        Load residues - amino acids.

        Returns:
            * PDBStream
        """
        from prody import parsePDBStream

        pdb_file = self.uniprot_dict[uniprot]

        with gunzipper(pdb_file) as pfile:
            aa = parsePDBStream(pfile)

        return aa

    def load_header(self, uniprot):
        """
        Load PDB header

        Returns:
            * dict
        """
        from prody import parsePDBHeader

        pdb_file = self.uniprot_dict[uniprot]
        return parsePDBHeader(pdb_file)

    def scrape_header(self, uniprot):
        """
        Scrape PDB header. Just gets alignment information for pdb.
        """
        import re
        import gzip

        header = dict()
        header[uniprot] = dict()
        
        with gzip.open(self.uniprot_dict[uniprot], mode='rt') as pfile:
            for line in pfile.readlines():
                if line[:5] == "DBREF":
                    l = line.strip()
                    l = re.split("\s+", l, flags=re.UNICODE)
                    header[uniprot]["db_first_from"] = int(l[3])
                    header[uniprot]["db_first_to"] = int(l[8])
                    header[uniprot]["db_accession"] = l[6]
                    header[uniprot]["db_ref"] = l[5].replace("UNP","UniProt")

        return header

    def load_rd(self, uniprot, return_model_confidence=False):
        """
        Load residue dict
        """
        aa = self.load(uniprot)
        res_map = dict(zip(aa.getResnums(),aa.getResnames()))
        rd = {k:v for k,v in res_map.items() if v in AMINO_ACID_MAP}

        if return_model_confidence:
            xx = aa.getResnums()
            bb = aa.getBetas()
            model_confidence = {}
            for idx,num in enumerate(xx):
                model_confidence[num] = bb[idx]
            return rd, model_confidence
        else:
            return rd

    def load_dm(self, uniprot, point='centroid', return_centroid_coord=False):
        """
        Load distance matrix for pdb.
        """
        from scipy.spatial.distance import euclidean
        import numpy as np

        aa = self.load(uniprot)

        xx = aa.getResnums()
        yy = aa.getCoords()
        zz = aa.getResnames()

        # Model confidence
        bb = aa.getBetas()

        pdb_resids = {}
        for i in range(len(xx)):
            if zz[i] in AMINO_ACID_MAP:
                pdb_resids[xx[i]] = True

        mapped_pdb_to_aa = defaultdict(set)
        model_confidence = defaultdict(float)

        for idx,resnum in enumerate(xx):
            if resnum in pdb_resids:
                mapped_pdb_to_aa[resnum].add(zz[idx])
                model_confidence[resnum] = bb[idx]

        coords = {}
        for i in range(len(xx)):
            if xx[i] not in pdb_resids:
                continue
            if xx[i] not in coords:
                coords[xx[i]] = []
            coords[xx[i]].append(yy[i])  ## add coordinates of an atom belonging to this residue

        ## Euclidean distance matrix
        D = []
        for i in range(len(pdb_resids)):
            D.append(np.zeros(i, dtype=np.float32))

        centroids = {}
        for k in coords:
            centroids[k] = np.mean(np.array(coords[k]), 0)

        co = [centroids[i] for i in pdb_resids]  ## pdb residue coordinates

        for i in range(len(pdb_resids)):
            for j in range(i):
                D[i][j] = euclidean(co[i], co[j])

        if return_centroid_coord:
            return (D, pdb_resids, mapped_pdb_to_aa, model_confidence, co)
        else:
            return (D, pdb_resids, mapped_pdb_to_aa, model_confidence)

class AccessionNo(object):
    """
    Accession Number Object.
    ---------------
    Stores Accession Number -> PDB mapping.
    """
    def __init__(
        self,
        accession_number,
        blast_dir="ref/refseq_blasted",
        pdb_dir="../../../getzlab-CLUMPS2/clumps/db/ref/pdbs"
    ):
        """
        Args:
            * x
        """
        xpo_param=(3,4.5,6,8,10)

        self.accession_number = accession_number
        self.blast_dir = blast_dir
        self.pdb_dir = pdb_dir

        # Get Blasted PDB
        self.blast = clumpsptm.mp.Blast(os.path.join(self.blast_dir,"{}.blasted.seq.gz".format(self.accession_number)))
        self.hit = self.blast.hits.iloc[0]
        self.pdb,self.chain = self.hit['Hit_def'].split(' ')[0].split("_")

        # Get PDB Info
        pdbstore = clumpsptm.PdbStore(self.pdb_dir)
        self.D,self.x,self.pdb_resnames = pdbstore.load_dm(self.pdb,self.chain)
        self.DDt = clumpsptm.transform_dx(self.D, xpo_param)

        # Align PDB Fasta to PDB SEQATOMS
        self.pdb_structure_idx = np.arange(max(list(self.pdb_resnames.keys())))+1
        self.pdb_structure_res = np.array(["."]*pdb_idx.shape[0])

        for idx in self.pdb_structure_idx-1:
            pdb_i = self.pdb_structure_idx[idx]
            try:
                self.pdb_structure_res[idx] = AMINO_ACID_MAP[list(self.pdb_resnames[pdb_i])[0]]
            except:
                pass

        self.pdb_structure_res = ''.join(self.pdb_structure_res)
        self.pdb_fasta_alignment = pairwise2.align.globalms(
            self.hit['Hsp_hseq'],
            self.pdb_structure_res.replace(".",""),1, -1, -.5, -.1
        )[0]
        self.offset = self.get_init_pdb_struct_i(self.pdb_structure_res) - self.get_alignment_offset(self.pdb_fasta_alignment[1])

    def get_init_pdb_struct_i(self, pdb_structure_res):
        """
        Get index of first registered atom.
        """
        for idx,a in enumerate(pdb_structure_res):
            if a!='.':
                return idx+1

    def get_alignment_offset(self, align):
        """
        Get offset from Fasta via alignment
        """
        c = 0
        for a in align:
            if a=='-':
                c+=1
            else:
                return c



# Modules
from . import mapping as mp
from . import vis as vis
from .pdbstore import PdbStore, AlphaStore
from .samplers import PTMSampler, PTMAlphaSampler
from .clumps import *

__version__ = '0.0.1'



import pandas as pd
import argparse
import numpy as np
import os
from .clumps import transform_dx, init_alg, clumps
from .pdbstore import PdbStore, AlphaStore
from .samplers import PTMSampler, PTMAlphaSampler
from .utils import generate_clumpsptm_output
from agutil.parallel import parallelize2

def main():
    parser = argparse.ArgumentParser(description='Run CLUMPS-PTM.')
    parser.add_argument('-i', '--input', required=True, help='<Required> Input file.')
    parser.add_argument('-m','--maps', required=True, help='<Required> Mapping with index as indices that overlap input.')
    parser.add_argument('-w', '--weight', required=True, help='<Required> Weighting for CLUMPS-PTM (ex. logFC).')
    parser.add_argument('-s','--pdbstore',required=True, help='<Required> path to PDBStore directory.')
    parser.add_argument('-o','--output_dir', default=".", help='Output directory.')
    parser.add_argument('-x', '--xpo', default=[3, 4.5, 6, 8, 10], type=list,
        help='Soft threshold parameter for truncated Gaussian.')
    parser.add_argument('--threads', type=int, default=1, help='Number of threads for sampling.')
    parser.add_argument('-v', '--verbose', action='store_true', default=False, help='Verbosity.')
    parser.add_argument('-f', '--features', nargs="*", default=None, help='Assays to subset for.')
    parser.add_argument('-g', '--grouping', default=None, help='DE group to use.')
    parser.add_argument('-q', '--use_only_significant_sites', action='store_true', help='Only use significant sites for CLUMPS-PTM.')
    parser.add_argument('--min_sites', default=3, help='Minimum number of sites.')
    parser.add_argument('--subset', default=None, help='Subset sites.', choices=('positive','negative'))
    parser.add_argument('--protein_id', default="accession_number", help='Unique protein id in input.')
    parser.add_argument('--site_id', default="variableSites", help='Unique site id in input.')
    parser.add_argument('--alphafold', action='store_true', default=False, help='Run using alphafold structures.')
    parser.add_argument('--alphafold_threshold', type=float, default=75, help='Threshold confidence level for alphafold sites.')
    args = parser.parse_args()

    print("---------------------------------------------------------")
    print("------------------ C L U M P S - P T M ------------------")
    print("---------------------------------------------------------")

    # ----------------------------------
    # Load files
    # ----------------------------------
    mapped_sites_df = pd.read_csv(args.maps, sep='\t', index_col=0)
    os.makedirs(args.output_dir, exist_ok=True)
    os.makedirs(os.path.join(args.output_dir, "clumpsptm"), exist_ok=True)

    if args.verbose:
        print("   * creating output directory: {}".format(args.output_dir))

    if args.verbose:
        print("   * loading {} mapped sites".format(mapped_sites_df.shape[0]))

    # Load PDB Store
    if args.alphafold:
        pdbstore = AlphaStore(args.pdbstore)
    else:
        pdbstore = PdbStore(args.pdbstore)

    # Load input file
    if args.input.endswith(".parquet"):
        de_df = pd.read_parquet(args.input)
        try:
            de_df = de_df.set_index("index")
        except:
            pass
    else:
        de_df = pd.read_csv(args.input, sep='\t')

        if 'index' in list(de_df):
            de_df = de_df.set_index("index")
        else:
            de_df = de_df.set_index(list(de_df)[0])

    if args.grouping is not None:
        de_df['id'] = de_df['id'].astype(str)
        de_df = de_df[de_df['id']==str(args.grouping)]
    if args.features is not None:
        de_df = de_df[de_df['feature'].isin(args.features)]

    # Combined Input with Mapping File
    de_df = mapped_sites_df.loc[np.intersect1d(mapped_sites_df.index, de_df.index)].join(
        de_df.drop(columns=np.intersect1d(de_df.columns, mapped_sites_df.columns))
    )

    # Subset for pos/neg sites
    if args.subset is not None:
        if args.verbose:
            print("   * subsetting for {} sites".format(args.subset))

        if args.subset == "positive":
            de_df = de_df[de_df[args.weight]>0]
        elif args.subset == "negative":
            de_df = de_df[de_df[args.weight]<0]
            de_df[args.weight] = -1 * de_df[args.weight]

    # Subset for significant sites
    # TODO: allow to select between p-value, fdr, and select threshold
    if args.use_only_significant_sites:
        print("   * filtering for significant sites below < 0.1 nominal-pvalue")
        de_df = de_df[de_df['P.Value']<0.1]

    if args.alphafold:
        print("   * filtering for sites in residues with > {} % alphafold prediction confidence.".format(args.alphafold_threshold))
        de_df = de_df[de_df['alphafold_confidence'] >= args.alphafold_threshold]

    # Subset for min sites
    gb = de_df.groupby(args.protein_id).size()
    de_df = de_df[de_df[args.protein_id].isin(gb[gb>=args.min_sites].index)]
    de_df['pdb_res_i'] = de_df['pdb_res_i'].astype(int)

    if args.verbose:
        print("   * {} input sites mapped".format(de_df.shape[0]))
        print("   * {} proteins with > {} sites".format(np.unique(de_df[args.protein_id]).shape[0], args.min_sites))

    # ----------------------------------
    # Run Clumps
    # ----------------------------------
    @parallelize2(maximum=args.threads)
    def run_clumps(acc):
        """
        Run CLUMPS PTM.
        """
        inputs_df = de_df[de_df[args.protein_id]==acc].copy()
        inputs_df = inputs_df.sort_values('pdb_res_i')

        if args.alphafold:
            uniprot = inputs_df.iloc[0]['uniprot']
        else:
            pdb, chain = inputs_df.iloc[0][['pdb','chain']]

        # Rule for multi-site modifications
        inputs_df = inputs_df.sort_values(['pdb_res_i', args.weight], ascending=[True, False]).drop_duplicates('pdb_res_i')

        # Get PDB Information
        if args.alphafold:
            D,x,pdb_resnames,model_confidence = pdbstore.load_dm(uniprot)
        else:
            D,x,pdb_resnames = pdbstore.load_dm(pdb, chain)

        DDt = transform_dx(D, args.xpo)

        # Get matched numerical site
        pdb_struct_i_to_d_map = {x:i for i,x in enumerate(np.sort(list(x.keys())))}
        inputs_df['d'] = inputs_df['pdb_res_i'].apply(lambda x: pdb_struct_i_to_d_map[int(x)])
        inputs_df['value'] = inputs_df[args.weight]

        def _run(feature):
            """Run."""
            if feature == "acetylome":
                _inputs_df = inputs_df[inputs_df["feature"]==feature].copy()
                _res = ["LYS"]
            elif feature == "phosphoproteome":
                _inputs_df = inputs_df[inputs_df["feature"]==feature].copy()
                _res = ["SER","THR","TYR"]
            elif feature == "ptm":
                _inputs_df = inputs_df.copy()
                _res = ["LYS","SER","THR","TYR"]

            if _inputs_df.shape[0] ==0:
                # print("     > {} - {} | p-value: {}".format(acc, feature, "no sites."))
                return

            # Initialize
            init_dict = init_alg(_inputs_df, DDt)

            if args.alphafold:
                sam = PTMAlphaSampler(
                    pdb_struct_i_to_d_map,
                    pdb_resnames,
                    model_confidence,
                    res=_res,
                    model_threshold=args.alphafold_threshold
                )
            else:
                sam = PTMSampler(pdb_struct_i_to_d_map, pdb_resnames, res=_res)

            # Run CLUMPS
            pval, wap_scr, niter = clumps(init_dict, sam, DDt, args.xpo, max_rand=1e4, use_booster=False)

            # Output
            # print("     > {} - {} | p-value: {}".format(acc, feature, sum(pval) / (5*niter)))
            _inputs_df.loc[:,"clumpsptm_niter"] = niter
            _inputs_df.loc[:,"clumpsptm_pval"] = sum(pval) / (5*niter)

            res_input = [str(x) for x in _inputs_df['pdb_res_i']]
            res_sample = [str(x) for x in sam.ptm]

            _inputs_df.loc[:,"clumpsptm_input"] = "+".join(res_input)
            _inputs_df.loc[:,"clumpsptm_sample"] = "+".join(res_sample)
            _inputs_df.loc[:,"clumpsptm_input_n"] = len(res_input)
            _inputs_df.loc[:,"clumpsptm_sample_n"] = len(res_sample)
            _inputs_df.to_parquet(os.path.join(args.output_dir, "clumpsptm", "{}_{}_clumpsptm.parquet".format(acc, feature)))

        if "acetylome" in args.features:
            try:
                _run("acetylome")
            except:
                print("     > ERROR for {} - {}".format(acc, "acetylome"))
        if "phosphoproteome" in args.features:
            try:
                _run("phosphoproteome")
            except:
                print("     > ERROR for {} - {}".format(acc, "phosphoproteome"))
        if "acetylome" in args.features and "phosphoproteome" in args.features:
            try:
                _run("ptm")
            except:
                print("     > ERROR for {} - {}".format(acc, "ptm"))

    print("   * running using {} threads".format(args.threads))

    # TODO: CHANGE
    protein_ids = np.unique(de_df[args.protein_id])

    tmp = [run_clumps(prot) for prot in protein_ids]
    results = [callback() for callback in tmp]

    # ----------------------------------
    # Generate Output File
    # ----------------------------------
    results_df = generate_clumpsptm_output(
        os.path.join(args.output_dir, "clumpsptm"),
        args.protein_id,
        args.site_id,
        alphafold=args.alphafold
    )
    results_df.to_csv(os.path.join(args.output_dir, "clumpsptm_output.tsv"), sep="\t")

if __name__ == "__main__":
    main()



from agutil.parallel import parallelize2
import os
import subprocess
from typing import Union

def blast_sequences(
    db_seq_file: str,
    blast_files: Union[list,str],
    output_dir: Union[None, str] = None,
    db_title: str = "ClumpsRefSeq",
    blast_dir_name = "blast_dir",
    seq_db_name = "seq_db",
    verbose: bool = True,
    n_threads: int = 15,
    ):
    """
    Blast Sequences
    --------------------------
    First, install blast:
        sudo apt-get install ncbi-blast+

    Args:
        * db_seq_file: fasta file to use for database with sequences of inputs
            * ex. PDB sequences, Uniprot
        * blast_files: file(s) of sequences to blast against
        * output_dir: directory to write outputs to
        * db_title: blast database title
        * verbose
        * n_threads: n_threads for blasts

    """
    # Make Blast DIrectory
    os.makedirs(os.path.join(output_dir, blast_dir_name), exist_ok=True)

    # Make Blast Database
    seq_db = os.path.join(output_dir, seq_db_name)
    os.makedirs(seq_db, exist_ok=True)

    # Make Blast Database
    cmd = "makeblastdb -in {} -dbtype 'prot' -out {}/{}/{} -title {}".format(db_seq_file, output_dir, seq_db_name, seq_db_name, db_title)
    db_out = subprocess.run(cmd, executable='/bin/bash', shell=True, stdout=subprocess.PIPE).stdout.decode()

    if verbose: print(db_out)

    @parallelize2(maximum=n_threads)
    def blast(seq):
        """Blast Command."""
        accession_no = seq.split('/')[-1].split('.seq')[0]

        cmd = "blastp -query {} -db {} -out {} -outfmt 5 && gzip -f {} ;".format(
            seq,
            os.path.join(output_dir, seq_db_name, seq_db_name),
            os.path.join(os.path.join(output_dir, blast_dir_name), "{}.blasted.seq".format(accession_no)),
            os.path.join(os.path.join(output_dir, blast_dir_name), "{}.blasted.seq".format(accession_no))
        )
        subprocess.call(cmd, executable='/bin/bash', shell=True)

    if verbose: print("   * Running {} blast sequences using {} threads".format(len(blast_files), n_threads))

    tmp = [blast(s) for s in blast_files]
    results = [callback() for callback in tmp]

class Blast(object):
    """
    Blast results
    ------------------
    This object store blast results.
    """
    def __init__(self, file, eval_thresh=0.001, ident_thresh=0):
        """
        Args:
            * file: blasted filename
            * eval_thresh: threshold for entries
            * ident_thresh: threshold of identity

        """
        from lxml import etree
        import gzip
        import pandas as pd

        self.file_name = file
        self.hits = list()
        self.list = list()
        self.idents = list()

        with gzip.open(file, 'r') as f:
            cur = etree.fromstring(f.read())
            qlen = 0  ## query length

            for i in cur:
                if i.tag == 'BlastOutput_iterations':
                    cur = i
                elif i.tag == 'BlastOutput_query-len':
                    qlen = int(i.text)

            cur = cur[-1]  ## last iteration

            for i in cur:
                if i.tag == 'Iteration_hits':
                    cur = i
                    break

            for hit in cur:
                hit_def = None
                hsps = None

                for i in hit:
                    if i.tag == 'Hit_def':
                        hit_def = i.text
                    elif i.tag == 'Hit_hsps':
                        hsps = i

                for hsp in hsps:
                    di = {}
                    di['Hit_def'] = hit_def

                    for i in hsp:
                        di[i.tag] = i.text

                    if float(di['Hsp_evalue']) > eval_thresh:  ### !!! FILTER !!!
                        continue

                    rel_ident = float(di['Hsp_identity'])/float(di['Hsp_align-len'])

                    if rel_ident < ident_thresh:         ### !!! FILTER !!!
                        continue

                    di['relIdentity'] = rel_ident
                    self.hits.append(di)
                    self.list.append(di['Hit_def'])

        self.hits = pd.DataFrame(self.hits)

    @property
    def nhits(self):
        return len(self.hits)

    @property
    def query(self):
        return self.file_name.split('/')[-1].split('.blasted')[0]

    def __str__(self):
        return "Blastp+ for {} | {} hits".format(self.query, self.nhits)

def load_accession_blast(a_num, unique_sites_df):
    """
    a_num
    """
    an_dict = {}
    an_dict['accession_no'] = a_num

    if scraped_accessions[a_num] is not None:
        an_dict['accession_no_matched'] = scraped_accessions[a_num].split('/')[-1].split('.seq')[0]

        uns = unique_sites_df.loc[a_num,'site_position']

        if isinstance(uns,np.int64):
            an_dict['acetyl_sites'] = np.array([uns,])
        else:
            an_dict['acetyl_sites'] = np.array(uns)

        an_dict['n_acetyl_sites'] = an_dict['acetyl_sites'].shape[0]

        matched = load_blasted_refseq(os.path.join(output_dir, "refseq_blasted", "{}.blasted.seq.gz".format(an_dict['accession_no_matched'])))

        if len(matched) == 0:
            return an_dict

        # Add PDB Chain & Name
        pdb_name = matched[0]['Hit_def'].split(' ')[0]
        an_dict['pdb'] = pdb_name.split('_')[0]
        an_dict['chain'] = pdb_name.split('_')[1]
        an_dict['pdb_desc'] = ' '.join(matched[0]['Hit_def'].split(' ')[1:])
        an_dict['pdb_ident'] = matched[0]['relIdentity']

        # Map Residues
        arr = np.array(list(matched[0]['Hsp_qseq']))

        # Get Overlapping Acetyl-Sites
        overlapping_sites = list()
        res_matches = list()

        for ac in an_dict['acetyl_sites']:
            try:
                res_match = arr[ac-int(matched[0]['Hsp_query-from'])]
            except:
                res_match = None

            if res_match=='K':
                overlapping_sites.append(ac)

            res_matches.append(res_match)

        an_dict['pdb_site_matches'] = overlapping_sites
        an_dict['pdb_n_site_matches']  = len(overlapping_sites)
        an_dict['Hsp_query-from'] = matched[0]['Hsp_query-from']
        an_dict['Hsp_query-to'] = matched[0]['Hsp_query-to']
        an_dict['Hsp_hit-from'] = matched[0]['Hsp_hit-from']
        an_dict['Hsp_hit-to'] = matched[0]['Hsp_hit-to']

        try:
            an_dict['percent_k'] = res_matches.count("K") / (len(an_dict['acetyl_sites'])-res_matches.count(None))
        except:
            an_dict['percent_k'] = None

    return an_dict



import sys
import os
import pandas as pd
import numpy as np

def transform_dx(D, xpo_param):
    """
    Transform distance matrix.
    --------------------------

    """
    import numpy as np

    DDt = []
    for soft_thresh_idx in xpo_param:
        den = 2.0 * soft_thresh_idx**2
        m = []
        for i in range(len(D)):
            mrow = np.zeros(i, dtype=np.float32)
            for j in range(i):
                mrow[j] = np.exp(-(D[i][j]**2)/den)
            m.append(mrow)
        DDt.append(m)

    return DDt

def wap(mut_indices, mvcorr, Mmv, DDt):
    """
    WAP Score
    --------------------------
    Compute WAP score to summarize pairwise distances between
    mutated residues in 3D protein structure.
    """
    import numpy as np

    s = np.zeros(len(DDt), np.float64)
    for mat in range(len(DDt)):
        d = DDt[mat]
        for i in range(len(mut_indices)):
            dcol = d[mut_indices[i]]
            for j in range(i):
                s[mat] += Mmv[mvcorr[i]][mvcorr[j]] * dcol[mut_indices[j]]
    return s

def init_alg(clumps_input_df, DDt):
    """
    Initialize CLUMPS algorithm.
    """
    mi = np.array(clumps_input_df['d'])
    mv = np.array(clumps_input_df['value'])

    # Compute matrix
    Mmv = []
    mvcorr = range(len(mv))

    for i in range(len(mi)):
        mrow = np.zeros(len(mi), np.float64)
        for j in range(len(mi)):
            mrow[j] = mv[i]*mv[j]
        Mmv.append(mrow)

    # Compute WAP score
    wap_obs = wap(mi, mvcorr, Mmv, DDt)

    # Init Params
    init_dict = {}
    init_dict['mi'] = mi
    init_dict['mv'] = mv
    init_dict['Mmv'] = Mmv
    init_dict['mvcorr'] = mvcorr
    init_dict['wap_obs'] = wap_obs

    return init_dict



from typing import Union
import glob
import os
from sys import stdout
import subprocess
from tqdm import tqdm
from collections import defaultdict
from agutil.parallel import parallelize2
import pandas as pd

from .blast import Blast
from ..utils import AMINO_ACID_MAP
from ..pdbstore import AlphaStore

def dl_ref(
    out_dir: str,
    files: Union[tuple, None] = None,
    verbose: bool = True
    ):
    """
    Download reference files
    --------------------------
    Downloads references files to provided directory.
       * PDB Seq Res (for Database)
       * RefSeq Latest Identifiers (fasta)

    """
    os.makedirs(out_dir, exist_ok=True)

    if files is None:
        files = (
            "ftp://ftp.wwpdb.org/pub/pdb/derived_data/pdb_seqres.txt",
            "ftp://ftp.ncbi.nlm.nih.gov/refseq/H_sapiens/annotation/GRCh38_latest/refseq_identifiers/GRCh38_latest_protein.faa.gz",
        )

    for file in files:
        if verbose:
            stdout.write("   * Downloading {}...\n".format(file.split('/')[-1]))
        cmd = "wget -P {} {}".format(out_dir, file)
        subprocess.call(cmd, executable='/bin/bash', shell=True)

        if file.endswith('.gz'):
            cmd = "gunzip -c {} > {}".format(os.path.join(out_dir, file.split('/')[-1]),os.path.join(out_dir, file.split('/')[-1].replace('.gz','')))
            subprocess.call(cmd, executable='/bin/bash', shell=True)

def split_fastas(
    fasta: str,
    out_dir: str,
    naming: str = "cptac"
    ):
    """
    Split fastas into individual files.
    """
    os.makedirs(out_dir, exist_ok=True)
    errors = list()

    with open(fasta, 'r') as f:
        data = f.read().split('>')

        for entry in tqdm(data[1:]):
            if naming=="cptac":
                f_name = entry.split(' ')[0]+'.seq'
            elif naming=="gencode":
                f_name = entry.split('|')[0]+'.seq'
            elif naming=="uniprot":
                try:
                    f_name = entry.split('|')[1]+'.seq'
                except:
                    errors.append(entry.split("\n")[0])
                    continue

            with open(os.path.join(out_dir, f_name), 'w') as f:
                f.write('>' + entry)

    if naming == "uniprot":
        print("   * {} sequences not parsed".format(len(errors)))
        return errors

def get_blast_hits_with_sifts(
    blasted_files,
    sifts_df,
    filter_only_human = False
    ):
    """
    Takes a list of blasted files, selects top blasted hit cross-references input
    Sifts data-base file.

    Parameters:
    -----------
    blasted_files [ required ]
        type: list

    sifts_db [ required ]
        type: pd.DataFrame

    filter_only_human [ optional ]
        default: False
        type: bool

    Returns:
    --------
    pd.DataFrame
    """
    import pandas as pd

    uniprots = set(sifts_df['SP_PRIMARY'])

    res = list()

    for bf in tqdm(blasted_files):
        bl = Blast(bf)

        try:
            bl.hits.loc[:,'uniprot'] = bl.hits['Hit_def'].apply(lambda x: x.split("|")[1])
            bl.hits.loc[:,'species'] = bl.hits['Hit_def'].apply(lambda x: x.split("|")[2].split(' ')[0].split("_")[1])
            bl.hits.loc[:,'query'] = bl.query

            if filter_only_human:
                bl.hits = bl.hits[bl.hits['species']=='HUMAN']

            hits_in_sifts = bl.hits[bl.hits['uniprot'].isin(uniprots)]

            if hits_in_sifts.shape[0] == 0:
                _res = bl.hits.iloc[0].copy()
                _res['sifts'] = False
            else:
                _res = hits_in_sifts.iloc[0].copy()
                _res['sifts'] = True

            _res['blast'] = True
        except:
            _res = pd.Series()
            _res['query'] = bl.query
            _res['blast'] = False

        res.append(_res)

    return pd.concat(res,1).T.set_index("query")

def get_blast_hits_with_alphafold(
    blasted_files: list,
    alphafold_dir: str,
    filter_only_human: bool = False
    ):
    """
    Takes a list of blasted files, selects top blasted hit cross-referenced
    with alpha-fold models.

    Parameters:
    -----------
    blasted_files [ required ]
        type: list

    alphafold_dir [ required ]
        type: str

    filter_only_human [ optional ]
        default: False
        type: bool

    Returns:
    --------
    pd.DataFrame
    """
    import pandas as pd

    alphaStore = AlphaStore(alphafold_dir)
    uniprots = set(alphaStore.uniprots)

    res = list()

    for bf in tqdm(blasted_files):
        bl = Blast(bf)

        try:
            bl.hits.loc[:,'uniprot'] = bl.hits['Hit_def'].apply(lambda x: x.split("|")[1])
            bl.hits.loc[:,'species'] = bl.hits['Hit_def'].apply(lambda x: x.split("|")[2].split(' ')[0].split("_")[1])
            bl.hits.loc[:,'query'] = bl.query

            if filter_only_human:
                bl.hits = bl.hits[bl.hits['species']=='HUMAN']

            hits_in_alphafold = bl.hits[bl.hits['uniprot'].isin(uniprots)]

            if hits_in_alphafold.shape[0] == 0:
                _res = bl.hits.iloc[0].copy()
                _res['alphafold'] = False
            else:
                _res = hits_in_alphafold.iloc[0].copy()
                _res['alphafold'] = True

            _res['blast'] = True
        except:
            _res = pd.Series()
            _res['query'] = bl.query
            _res['blast'] = False

        res.append(_res)

    return pd.concat(res,1).T.set_index("query")

def get_pdb_match(sifts_s, acc_df, pdbstore):
    """
    For a given sifts entry, this loads and concatenates PDB information.

    Parameters:
    -----------
    sifts_s [ required ]
        type: pd.Series

    acc_df [ required ]
        type: pd.DataFrame

    pdbstore [ required ]
        type: clumpsptm.pdbstore.PdbStore

    Returns:
    --------
    pd.DataFrame
        accession numbers as indices and corresponding PDB res_num & res
    """
    import numpy as np

    pdb_resnames = pdbstore.load_rd(sifts_s["PDB"], sifts_s["CHAIN"])

    def _try_map(x):
        try:
            return AMINO_ACID_MAP[pdb_resnames[x]]
        except:
            return None

    # PDB IDs
    pdb_i = np.unique(list(pdb_resnames.keys()))

    # Uniprot 2 PDB
    md = defaultdict(int,{x+sifts_s['db_first_to']-sifts_s['db_first_from']:x for x in pdb_i})

    _pdb_res = '{}_{}_res'.format(sifts_s['PDB'], sifts_s['CHAIN'])
    _pdb_res_i = '{}_{}_res_i'.format(sifts_s['PDB'], sifts_s['CHAIN'])
    _pdb_res_match = '{}_{}_res_match'.format(sifts_s['PDB'], sifts_s['CHAIN'])

    acc_df[_pdb_res_i] = acc_df['uniprot_res_i'].apply(lambda x: md[x])
    acc_df[_pdb_res] = acc_df[_pdb_res_i].apply(_try_map, 1)
    acc_df[_pdb_res_match] = acc_df['acc_res']==acc_df[_pdb_res]

    return acc_df[[_pdb_res_i,_pdb_res,_pdb_res_match]]

def get_pdb_matches(
    proteins,
    ptm_df,
    sifts_df,
    pdbstore,
    out_dir,
    n_threads,
    protein_id = "accession_number"
    ):
    """
    For a list of proteins, use multi-threading to curate mapped PDB
    structures using SIFTS database.

    Parameters:
    -----------
    proteins [ required ]
        type: iterable

    ptm_df [ required ]
        type: pd.DataFrame

    sifts_df [ required ]
        type: pd.DataFrame

    pdbstore [ required ]
        type: clumpsptm.pdbstore.PdbStore

    out_dir [ required ]
        type: str

    n_threads [ required ]
        type: int

    protein_id [ optional ]
        default: accession_number
        type: str

    Returns:
    --------
    set
        Accession numbers with errors in mapping.
    """
    # Make directories
    os.makedirs(out_dir, exist_ok=True)
    os.makedirs(os.path.join(out_dir, "acc_mapped_sites"), exist_ok=True)
    os.makedirs(os.path.join(out_dir, "acc_mapped_sifts"), exist_ok=True)

    # Multi-threading function
    @parallelize2(maximum=n_threads)
    def get_pdb_match_caller(acc):
        """Func for matching"""
        try:
            acc_df = ptm_df[ptm_df[protein_id]==acc].copy().sort_values("uniprot_res_i")[[
                'uniprot',
                'acc_res_i',
                'acc_res_idx',
                'acc_res',
                'uniprot_res_i',
                'uniprot_res',
                'uniprot_match'
            ]]

            uniprot = acc_df.iloc[0]['uniprot']

            sifts_filt_df = sifts_df[sifts_df['SP_PRIMARY']==uniprot].copy().sort_values(['SP_BEG','len'], ascending=[True,False])
            sifts_filt_df = sifts_filt_df.drop_duplicates(subset=['PDB','CHAIN'])

            resd = list()

            for idx,row in sifts_filt_df.iterrows():
                try:
                    _resd = get_pdb_match(row, acc_df, pdbstore)
                    resd.append(_resd)
                    sifts_filt_df.loc[idx,'percent_match'] = 100 * sum(_resd.iloc[:,-1]) / _resd.shape[0]
                except:
                    print("Error loading: {}-{}".format(row["PDB"], row["CHAIN"]))

            resd = pd.concat(resd,1)
            sifts_filt_df['proteins'] = acc
            sifts_filt_df = sifts_filt_df.astype(str)

            resd.to_parquet(os.path.join(out_dir, "acc_mapped_sites", "{}.parquet".format(acc)))
            sifts_filt_df.to_parquet(os.path.join(out_dir, "acc_mapped_sifts", "{}.parquet".format(acc)))

            return None
        except:
            return acc

    tmp = [get_pdb_match_caller(acc) for acc in proteins]
    err = {callback() for callback in tmp}

    return err

def get_pdb_matches_alphafold(
    proteins,
    ptm_df,
    alphaStore,
    out_dir,
    n_threads,
    protein_id: str = "accession_number"
    ):
    """
    For a list of proteins, use multi-threading to curate mapped PDB
    structures with overlapping residues from AlphaFold structures.

    Parameters:
    -----------
    proteins [ required ]
        type: iterable

    ptm_df [ required ]
        type: pd.DataFrame

    alphaStore [ required ]
        type: clumpsptm.pdbstore.AlphaStore

    out_dir [ required ]
        type: str

    n_threads [ required ]
        type: int

    protein_id [ optional ]
        default: accession_number
        type: str

    Returns:
    --------
    set
        Uniprots with errors in mapping.
    """
    # Make directories
    os.makedirs(out_dir, exist_ok=True)
    os.makedirs(os.path.join(out_dir, "alphafold_mapped_sites"), exist_ok=True)

    # Multi-threading function
    @parallelize2(maximum=n_threads)
    def get_pdb_match_caller(prot):
        """Func for matching"""
        try:
            prot_df = ptm_df[ptm_df[protein_id]==prot].copy().sort_values("uniprot_res_i")[[
                'uniprot',
                'acc_res_i',
                'acc_res_idx',
                'acc_res',
                'uniprot_res_i',
                'uniprot_res',
                'uniprot_match',
                'pdb_res_i'
            ]]

            uniprot = prot_df.iloc[0]['uniprot']
            rd, model_confidence = alphaStore.load_rd(uniprot, return_model_confidence=True)

            pdb_res_row = list()
            model_confidence_row = list()

            for idx,row in prot_df.iterrows():
                pdb_res = "X"
                model_conf = 0
                try:
                    pdb_res = AMINO_ACID_MAP[rd[row['pdb_res_i']]]
                    model_conf = model_confidence[row['pdb_res_i']]
                except:
                    pass

                pdb_res_row.append(pdb_res)
                model_confidence_row.append(model_conf)

            prot_df.loc[:,'pdb_res'] = pdb_res_row
            prot_df.loc[:,'alphafold_confidence'] = model_confidence_row
            prot_df.to_parquet(os.path.join(out_dir, "alphafold_mapped_sites", "{}.parquet".format(prot)))

            return None
        except:
            return prot

    tmp = [get_pdb_match_caller(prot) for prot in proteins]
    err = {callback() for callback in tmp}

    return err



from .blast import blast_sequences
from .blast import Blast
from .blast import load_accession_blast
from .ref import dl_ref, split_fastas, get_blast_hits_with_sifts, get_pdb_matches
from .ref import get_blast_hits_with_alphafold, get_pdb_matches_alphafold



import pandas as pd
import matplotlib.pyplot as plt
import numpy as np

def dotplot(
    results_df,
    y="y",
    x="clumpsptm_fdr",
    thresh=0.1,
    sort_by="ptm",
    ax=None,
    shape_dict = {'ptm':'o', 'phosphoproteome':',', 'acetylome':'^'},
    name_dict = {'ptm':'Combined', 'phosphoproteome':'Phosphorylation', 'acetylome':'Acetylation'},
    color_dict = {'fdr':'red', 'pval':'orange', 'nsig':'lightgrey'},
    title = None,
    show_gene_only = True,
    ):
    """
    Plot summaryplot for CLUMPS-PTM.
    """
    import matplotlib.patches as mpatches

    def _assign_color(row):
        if row['clumpsptm_fdr'] <= thresh:
            return color_dict['fdr']
        elif row['clumpsptm_fdr'] <= 0.25:
            return color_dict['pval']
        else:
            return color_dict['nsig']

    # Select Structures to Plot (y-axis)
    y_to_plot = np.unique(results_df[
        (results_df['clumpsptm_pval']<=thresh) &
        (results_df['clumpsptm_sampler']==sort_by)
        ][y])

    if y_to_plot.shape[0] == 0:
        print("No Nominal P-Values < 0.1")
        return

    # Subset for Data
    data_to_plot = results_df[results_df[y].isin(y_to_plot)].copy()
    assays = set(data_to_plot['clumpsptm_sampler']) - {sort_by}
    data_to_plot['category'] = pd.Categorical(data_to_plot['clumpsptm_sampler'], [sort_by]+list(assays))

    # Log FDR & color
    data_to_plot.loc[data_to_plot[data_to_plot[x]==0].index,x] = np.min(
        data_to_plot[x][data_to_plot[x]>0]*0.1)

    data_to_plot['x'] = -np.log10(data_to_plot[x])
    data_to_plot['c'] = data_to_plot.apply(_assign_color, 1)

    # Sort values
    data_to_plot = data_to_plot.sort_values(["category","x"], ascending=[True, True])

    # Create figure
    if ax is None:
        fig, ax = plt.subplots(figsize=(4, 2 + int(data_to_plot.shape[0]/20)))

    # Plot
    for idx, cat in enumerate(data_to_plot['category'].cat.categories):
        ax.scatter(
            y=y, x="x", c="c",
            data=data_to_plot[data_to_plot['clumpsptm_sampler']==cat],
            linewidth=1, edgecolor="black", zorder=10-idx,
            marker=shape_dict[cat], label=name_dict[cat], s=100, alpha=0.6
        )

    for j in range(y_to_plot.shape[0]):
        ax.axhline(j, c='lightgrey', alpha=0.4)

    ax.set_ylim(-1, y_to_plot.shape[0]+.5)
    ax.axvline(-np.log10(thresh), linewidth=0.5, c='lightgrey')
    ax.tick_params(axis='y', which='major', labelsize=7)

    if 'pval' in x:
        ax.set_xlabel("-log_10 P-Val")
    elif 'fdr' in x:
        ax.set_xlabel("-log10 FDR")

    handles, labels = ax.get_legend_handles_labels()

    # Patches
    patch = mpatches.Patch(
        facecolor=color_dict['fdr'],
        label="< {} {}".format(0.1, "FDR"),
        edgecolor='k',
        linewidth=1,
    )

    handles.append(patch)

    patch = mpatches.Patch(
        facecolor=color_dict['pval'],
        label="< {} {}".format(0.25, "FDR"),
        edgecolor='k',
        linewidth=1,
    )

    handles.append(patch)

    ax.legend(handles=handles, loc='lower left', bbox_to_anchor=(1, 0))

    if title is None:
        ax.set_title(results_df['id'][0])

    if show_gene_only and y=='y':
        plt.draw()
        labels = [item.get_text().split(" | ")[1] for item in ax.get_yticklabels()]
        ax.set_yticklabels(labels)



from typing import Union

from .funcs import wap

def clumps(
    init_dict,
    sam,
    DDt,
    xpo_param: Union[list,tuple],
    seed: int = 1,
    max_rand: int = 1e8,
    use_booster: bool = False,
    booster_interval: int = 1000,
    verbose: bool = False,
    ):
    """
    Run Clumps Algorithm
    ------------
    Returns:
        * p-values
        * wap score aggr
        * n-iterations
    """
    from sys import stdout
    import numpy as np

    def booster():
        """
        Implements the booster algorithm by Getz et al. that saves CPU time.
        Returns False if the randomization should be stopped.
        """
        ret = False
        for i in range(len(xpo_param)):
            s = (rnd - p[i] + 1.0) / ((rnd + 3)*(p[i] + 1))
            if s >= 0.05271:  ## =0.9/(2*1.96)]**2:
                ret = True
                break
        return ret

    if seed is not None:
        np.random.seed(seed)

    p = [0]*len(xpo_param)
    wap_rnd = [0]*len(xpo_param)
    rnd = 0

    for rnd in range(int(max_rand)):
        if rnd%booster_interval == 0 and rnd > 0 and use_booster:
            if verbose:
                stdout.write("\r{}/{}".format(rnd,int(max_rand)))
            if not booster():
                break

        # Randomly permute
        mi,mvcorr = sam.sample(init_dict['mi'])

        # Compute test WAP
        r = wap(mi, mvcorr, init_dict['Mmv'], DDt)

        for rr in range(len(xpo_param)):
            wap_rnd[rr] += r[rr]
            if r[rr] >= init_dict['wap_obs'][rr]:
                p[rr] += 1

    return p, wap_rnd, rnd+1



import numpy as np
import random

class PTMSampler(object):
    """
    Samples only modifiable residues.
    --------------------------
    Takes in pdb residue id (pr) and a mapping default dictionary.
    Samples only from residues that can be acetylated or phosphorylated.
    """
    def __init__(self, pdb_resid_to_d_map, pdb_resnames, res=['LYS','SER','THR','TYR']):
        self.resid_to_d_map = pdb_resid_to_d_map
        self.mapping = pdb_resnames
        self.ptm = list()

        for idx in pdb_resnames.keys():
            try:
                if list(pdb_resnames[idx])[0] in res:
                    self.ptm.append(idx)
            except:
                pass

    def sample(self, mi):
        sorted_indices = sorted(random.sample(self.ptm, len(mi)))
        random_permuts = np.random.permutation(len(mi))
        return [self.resid_to_d_map[x] for x in sorted_indices], random_permuts

class PTMAlphaSampler(object):
    """
    Samples only modifiable residues passing AlphaFold Confidence threshold.
    --------------------------
    Takes in pdb residue id (pr) and a mapping default dictionary.
    Samples only from residues that can be acetylated or phosphorylated.

    """
    def __init__(self, pdb_resid_to_d_map, pdb_resnames, model_confidence, res=['LYS','SER','THR','TYR'], model_threshold=75):
        self.resid_to_d_map = pdb_resid_to_d_map
        self.mapping = pdb_resnames
        self.confidence = model_confidence
        self.ptm = list()

        for idx in pdb_resnames.keys():
            try:
                if list(pdb_resnames[idx])[0] in res:
                    if model_confidence[idx] >= model_threshold:
                        self.ptm.append(idx)
            except:
                pass

    def sample(self, mi):
        sorted_indices = sorted(random.sample(self.ptm, len(mi)))
        random_permuts = np.random.permutation(len(mi))
        return [self.resid_to_d_map[x] for x in sorted_indices], random_permuts



from setuptools import setup
import sys

if sys.version_info < (3,6):
    import warnings
    print("This python version is not supported:")
    print(sys.version)
    print("CLUMPS-PTM requires python 3.6 or greater")
    sys.exit(1)

setup(
    name='clumps-ptm',
    author='Shankara Anand',
    version="0.0.2",
    author_email='sanand@broadinstitute.org',
    description='CLUMPS-PTM driver gene discovery using 3D protein structure (Getz Lab).',
    install_requires=[
        "matplotlib>=3.2.1",
        "twobitreader>=3.1.7",
        "statsmodels>=0.9.0",
        "scipy>=1.3.0",
        "pyopenssl>=19.0.0",
        "prody>=1.10.10",
        "lxml>=4.3.4",
        "biopython>=1.73",
        "tqdm>=4.32.1",
        "agutil"
    ],
    entry_points={
        'console_scripts':[
            'clumpsptm = clumpsptm.__main__:main',
        ]
    }
)
